#version 430 core

#define PCSS 0
#define MVSS 1
#define SHADOWS PCSS

uniform vec3 overlayColor;
uniform uint colorID;
uniform samplerCube environmentMap; // texture unit 1

// lighting
uniform vec3 ambientColor;

const int NUM_LIGHTS_MAX = 8;
const int NUM_SHADOWS_MAX = 4;
uniform int numLights;
uniform vec3[NUM_LIGHTS_MAX] lightViewPosition;
uniform vec3[NUM_LIGHTS_MAX] lightViewDirection;
uniform vec3[NUM_LIGHTS_MAX] lightColor;
uniform vec3[NUM_LIGHTS_MAX] lightAttenuation;
uniform vec3[NUM_LIGHTS_MAX] lightSpotFactors;
uniform int[NUM_LIGHTS_MAX] lightType;
uniform ivec2[NUM_LIGHTS_MAX] lightShadowMapIndex;
uniform mat4[NUM_SHADOWS_MAX] lightViewProjection;
uniform vec4[NUM_SHADOWS_MAX] lightDepthConversion;

// materials
subroutine vec3 fetchColorType();
layout(location=0) subroutine uniform fetchColorType fetchColor;

subroutine void fetchNVType(out vec3 n, out vec3 v);
layout(location=1) subroutine uniform fetchNVType fetchNV;

subroutine float fetchSpecularType();
layout(location=2) subroutine uniform fetchSpecularType fetchSpecular;

uniform sampler2DArray shadowMap; // texture unit 0

uniform vec4 reflectionFactors_shininess;
uniform sampler2D diffuseMap; // texture unit 2
uniform sampler2D normalMap; // texture unit 3
uniform sampler2D specularMap; // texture unit 4
uniform float normalMapStrength;


// camera data
uniform vec2 invImageSize; // 1.0/width, 1.0/height
uniform vec2 frustumSize; // frustum of directional light
uniform mat3 invViewMatrix3x3;

uniform vec4 dofDistances; // near start, near end, far start, far end
uniform vec2 dofBlurRadii; // max near blur radius, max far blur radius


in vec3 vf_normal;
in vec3 vf_tangent;
in vec3 vf_bitangent;
in vec3 vf_viewPosition;
in vec3 vf_worldPosition;
in vec3 vf_color;
in vec2 vf_texCoords;

out vec4 outColor;
out float outID;
out vec4 outDof_Depth;

vec2 poissonDisk[32] = vec2[](
	vec2(-0.613392, 0.617481),
	vec2(0.170019, -0.040254),
	vec2(-0.299417, 0.791925),
	vec2(0.645680, 0.493210),
	vec2(-0.651784, 0.717887),
	vec2(0.421003, 0.027070),
	vec2(-0.817194, -0.271096),
	vec2(-0.705374, -0.668203),
	vec2(0.977050, -0.108615),
	vec2(0.063326, 0.142369),
	vec2(0.203528, 0.214331),
	vec2(-0.667531, 0.326090),
	vec2(-0.098422, -0.295755),
	vec2(-0.885922, 0.215369),
	vec2(0.566637, 0.605213),
	vec2(0.039766, -0.396100),
	vec2(0.751946, 0.453352),
	vec2(0.078707, -0.715323),
	vec2(-0.075838, -0.529344),
	vec2(0.724479, -0.580798),
	vec2(0.222999, -0.215125),
	vec2(-0.467574, -0.405438),
	vec2(-0.248268, -0.814753),
	vec2(0.354411, -0.887570),
	vec2(0.175817, 0.382366),
	vec2(0.487472, -0.063082),
	vec2(-0.084078, 0.898312),
	vec2(0.488876, -0.783441),
	vec2(0.470016, 0.217933),
	vec2(-0.696890, -0.549791),
	vec2(-0.149693, 0.605762),
	vec2(0.034211, 0.979980)
);

float random(vec3 seed, int i){
	vec4 seed4 = vec4(seed,i);
	float dot_product = dot(seed4, vec4(12.9898,78.233,45.164,94.673));
	return fract(sin(dot_product) * 43758.5453);
}

float getVisibilityPCSS(vec3 shadowCoord, float bias, int shadowMapIndex)
{
	float LIGHT_SIZE_UV = 3.f / frustumSize.x; // 3.f is the light source's size in world coordinates
	float searchWidth = LIGHT_SIZE_UV;
	vec4 conv = lightDepthConversion[shadowMapIndex];

	float blockerSum = 0.f;
	float numBlockers = 0.f;

	for (int i=0; i<32; i++)
	{
		vec2 offset = poissonDisk[i]*searchWidth;
		float compare = shadowCoord.z - 0.0005f - bias * (offset.x+offset.y)/(12.f*invImageSize.x);
		float shadowMapDepth = texture(shadowMap, vec3(shadowCoord.xy + offset, shadowMapIndex)).r;
		float linearDepth = (conv.x + conv.y*shadowMapDepth) / (conv.z + conv.w*shadowMapDepth);

		float blocking = step(linearDepth,compare);

		blockerSum += blocking * linearDepth;
		numBlockers += blocking;
	}
	
	if (numBlockers == 0.f)
	{
		return 1.f;
	}
	else
	{
		float avgBlockerDepth = blockerSum / numBlockers;
		float penumbraRatio = shadowCoord.z - avgBlockerDepth;
		float filterRadius = 5.f * penumbraRatio * LIGHT_SIZE_UV;
		float sum = 0.0f; 
		for ( int i = 0; i < 32; i++ ) 
		{
			vec2 offset = poissonDisk[i] * filterRadius; 
			float compare = shadowCoord.z - 0.0005f - bias * length(offset)/(4.f*invImageSize.x);
			float shadowMapDepth = texture(shadowMap, vec3(shadowCoord.xy + offset, shadowMapIndex)).r;
			float linearDepth = (conv.x + conv.y*shadowMapDepth) / (conv.z + conv.w*shadowMapDepth);
			sum += float(linearDepth > compare);
		}
		return sum/32.f;
	}
}

float getVisibilityPCF(vec3 shadowCoord, float bias, int shadowMapIndex)
{
	float visibility = 1.0;

	float sum = 0.0f;
	vec2 factor = 0.3f/frustumSize;
	for (int i=0; i<4; i++)
	{
		int index = int(16.0*random(floor(vf_worldPosition.xyz*300.0), i))%16;
		vec2 offset=poissonDisk[index]*factor;
		float compare = shadowCoord.z - bias*length(offset)*200.f;
		bool smaller = texture( shadowMap, vec3(shadowCoord.xy + offset, shadowMapIndex) ).r < compare;
		sum += float(smaller);
	}
	if (sum == 0 || sum == 4) {
		return (4.f - sum) * 0.25f;
	}
	for (int i=4; i<16; i++)
	{
		int index = int(16.0*random(floor(vf_worldPosition.xyz*300.0), i))%16;
		vec2 offset=poissonDisk[index]*factor;
		float compare = shadowCoord.z - bias*length(offset)*200.f;
		bool smaller = texture( shadowMap, vec3(shadowCoord.xy + offset, shadowMapIndex) ).r < compare;
		sum += float(smaller);
	}
	visibility -= sum * (1.f/16.f);
	return visibility;
}

subroutine(fetchColorType)
vec3 fetchColorFromDiffuseMap()
{
	return texture(diffuseMap, vf_texCoords).xyz;
}

subroutine(fetchColorType)
vec3 fetchColorFromVertex()
{
	return vf_color;
}

subroutine(fetchNVType)
void fetchNVWithNormalMapping(out vec3 n, out vec3 v)
{
	mat3 tangentToViewSpace = mat3(
		normalize(vf_tangent),
		normalize(vf_bitangent),
		normalize(vf_normal)
	);
	n =	tangentToViewSpace * normalize(vec3(normalMapStrength,normalMapStrength,1.f) * (texture(normalMap, vf_texCoords).xyz*2.f - 1.f));
	v = normalize(vf_viewPosition);
}

subroutine(fetchNVType)
void fetchNVWithoutNormalMapping(out vec3 n, out vec3 v)
{
	n = normalize(vf_normal);
	v = normalize(vf_viewPosition);
}

subroutine(fetchSpecularType)
float fetchSpecularFromMap()
{
	return texture(specularMap, vf_texCoords).r;
}

subroutine(fetchSpecularType)
float fetchSpecular1()
{
	return 1.f;
}

void main(void)
{
	vec3 totalLight = ambientColor * reflectionFactors_shininess.x;
	vec3 n,v;
	fetchNV(n,v);

	//vec3 refl_v = invViewMatrix3x3 * reflect(v,n);
	//vec3 envMapColor = 2.f * texture(environmentMap, refl_v).xyz;

	vec3 baseColor = fetchColor();
	float specular = fetchSpecular();

	for (int i=0; i<numLights; i++)
	{
		if (lightType[i] == 0) // Point light
		{
			vec3 l = lightViewPosition[i] - vf_viewPosition; // point to light
			
			float dist = length(l); // distance point to light
			float attenuation = 1.f / (lightAttenuation[i].x + lightAttenuation[i].y*dist + lightAttenuation[i].z*dist*dist);

			// phong
			vec3 lNorm = l / dist;
			vec3 r = normalize(reflect(lNorm, n));
			float spec = max(pow(dot(v, r), reflectionFactors_shininess.w), 0);

			float cosTheta = clamp(dot(lNorm, n), 0.0, 1.0); // cos theta
			totalLight += lightColor[i] * (attenuation * reflectionFactors_shininess.y + spec * reflectionFactors_shininess.z*specular) * cosTheta;
		}
		else if (lightType[i] == 1) // Directional light
		{
			vec3 lNorm = -lightViewDirection[i];
			float cosTheta = clamp(dot(lNorm,n), 0.0, 1.0);

			float visibility = 1.0f;
			int shadowMapIndex = lightShadowMapIndex[i].x;
			if (shadowMapIndex != -1 && cosTheta > 0.01f)
			{
				float bias = 0.005*tan(acos(cosTheta)); // cosTheta is dot( n,l ), clamped between 0 and 1
				bias = clamp(bias, 0.001f, 0.01f);
				vec3 shadowCoord = (lightViewProjection[shadowMapIndex] * vec4(vf_worldPosition,1.0)).xyz;
				shadowCoord.xyz = shadowCoord.xyz*0.5f + 0.5f;
				//visibility = getVisibilityPCF(shadowCoord, bias, shadowMapIndex);
				visibility = getVisibilityPCSS(shadowCoord, bias, shadowMapIndex);
			}
			// phong
			vec3 r = normalize(reflect(lNorm, n));
			float spec = max(pow(dot(v, r), reflectionFactors_shininess.w), 0);
			totalLight += lightColor[i]* (reflectionFactors_shininess.y + spec*reflectionFactors_shininess.z*specular)*cosTheta*visibility;
		}
		else // Spot light
		{
			vec3 l = lightViewPosition[i] - vf_viewPosition; // point to light
			float dist = length(l); // distance point to light
			vec3 lNorm = l / dist; // direction from point to light

			// this is in view space
			float rho = dot(lNorm, -lightViewDirection[i]); // cos of angle between l and lightdirection
			float spot = pow( clamp( (rho - lightSpotFactors[i].x) / (lightSpotFactors[i].y - lightSpotFactors[i].x), 0.0f, 1.0f ), lightSpotFactors[i].z ); // spot factor	
			float attenuation = 1.f / (lightAttenuation[i].x + lightAttenuation[i].y*dist + lightAttenuation[i].z*dist*dist);

			float cosTheta = clamp(dot(lNorm, n), 0.0, 1.0); // cos theta

			float cosThetaTimesSpot = cosTheta * spot;

			// phong
			vec3 r = normalize(reflect(lNorm,n));
			float spec = max(pow(dot(v, r), reflectionFactors_shininess.w),0);

			float visibility = 1.0f;
			int shadowMapIndex = lightShadowMapIndex[i].x;
			if (shadowMapIndex != -1 && cosThetaTimesSpot > 0.01f)
			{
				if (SHADOWS == PCSS)
				{
					float bias = 0.005f*sqrt(1-cosTheta*cosTheta)/cosTheta;//0.005f*tan(acos(cosTheta)); // cosTheta is dot( n,l ), clamped between 0 and 1
					bias = clamp(bias, 0.001f,0.01f);
					vec4 shadowCoord = lightViewProjection[shadowMapIndex] * vec4(vf_worldPosition,1.0);
					vec4 conv = lightDepthConversion[shadowMapIndex];
					shadowCoord.xyz = shadowCoord.xyz / shadowCoord.w;
					shadowCoord.xyz = shadowCoord.xyz * 0.5f + 0.5f;
					shadowCoord.z = (conv.x + conv.y * shadowCoord.z) / (conv.z + conv.w * shadowCoord.z);
					visibility = getVisibilityPCSS(shadowCoord.xyz, bias, shadowMapIndex);
				}
				//visibility = getVisibilityPCF(shadowCoord.xyz, bias, shadowMapIndex);
			}

			totalLight += lightColor[i] * (attenuation * reflectionFactors_shininess.y + spec * reflectionFactors_shininess.z*specular) * cosThetaTimesSpot * visibility;
		}
	}

	float viewDepth = -vf_viewPosition.z;
	float coc = (1.f - smoothstep(dofDistances.x, dofDistances.y, viewDepth)) * dofBlurRadii.x
		+ smoothstep(dofDistances.z, dofDistances.w, viewDepth) * dofBlurRadii.y;
	coc = max(coc, 1.f);
	float intensity = 1.f/(coc*coc);

	outColor = vec4(overlayColor*baseColor*totalLight, 1.0);
	outDof_Depth = vec4(coc, intensity, 0.f, vf_viewPosition.z);
	outID = float(colorID) / 32767.f; // divide by (2^15 - 1)
}