#version 430 core

uniform sampler2D accumulationMap; // slot 0
uniform sampler2D diffuseMap; // slot 1
uniform sampler2D specularMap; // slot 2

uniform sampler2D normalMap; // slot 3, these values are normalized from [-1,1] -> [0,1]
uniform samplerCube environmentMap; // slot 4
uniform vec3 envBoxMin;
uniform vec3 envBoxMax;
uniform vec3 envBoxPosition;

uniform vec3 reflectionFactors; // diffuse factor, specular Factor, environment factor
uniform vec3 diffuseColor;
uniform vec3 ambientLight;

uniform mat3 invViewMatrix3x3;

uniform vec2 invImageSize;

subroutine vec3 fetchColorType();
layout(location=0) subroutine uniform fetchColorType fetchColor;

subroutine float fetchSpecularType();
layout(location=1) subroutine uniform fetchSpecularType fetchSpecular;

subroutine vec3 fetchEnvironmentColorType();
layout(location=2) subroutine uniform fetchEnvironmentColorType fetchEnvironmentColor;

in vec3 vf_worldPosition;
in vec3 vf_worldViewVector; // surface to viewer in world space
in vec2 vf_texCoords;
in vec4 vf_prevScreenPosition;
in vec4 vf_screenPosition;

layout(location=0) out vec3 outColor;
layout(location=1) out vec2 velocity;

#include<lum.glsl>

subroutine(fetchColorType)
vec3 fetchColorFromDiffuseMap()
{
	return texture(diffuseMap, vf_texCoords).xyz;
}

subroutine(fetchColorType)
vec3 fetchColorFromUniform()
{
	return diffuseColor;
}

subroutine(fetchSpecularType)
float fetchSpecularFromMap()
{
	return texture(specularMap, vf_texCoords).r;
}

subroutine(fetchSpecularType)
float fetchSpecular1()
{
	return 1.f;
}

subroutine(fetchEnvironmentColorType)
vec3 fetchEnvironmentColorFromMap()
{
	vec3 worldNormal = invViewMatrix3x3 * (2.f * texelFetch(normalMap, ivec2(gl_FragCoord.xy), 0).xyz - 1.f);
	vec3 worldReflectionVector = reflect(vf_worldViewVector, worldNormal);

	vec3 firstPlaneIntersect = (envBoxMax - vf_worldPosition) / worldReflectionVector;
	vec3 secondPlaneIntersect = (envBoxMin - vf_worldPosition) / worldReflectionVector;
	vec3 furthestPlane = max(firstPlaneIntersect, secondPlaneIntersect);
	float dist = min(min(furthestPlane.x, furthestPlane.y), furthestPlane.z);
	vec3 intersectionPoint = vf_worldPosition + worldReflectionVector * dist;
	worldReflectionVector = intersectionPoint - envBoxPosition;

	return texture(environmentMap, worldReflectionVector).rgb;
}

subroutine(fetchEnvironmentColorType)
vec3 fetchEnvironmentColorNone()
{
	return vec3(0.f);
}

void main(void)
{
	vec4 accumLight = texelFetch(accumulationMap, ivec2(gl_FragCoord.xy), 0);
	vec3 diffuseColor = fetchColor(); // vertex color or color from diffuse map
	float specularFactor = fetchSpecular(); // 1.f or value from specular map

	vec3 environmentColor = fetchEnvironmentColor();

	vec2 screenPosition = (vf_screenPosition.xy / vf_screenPosition.w)*0.5f+0.5f;
	vec2 prevScreenPosition = (vf_prevScreenPosition.xy / vf_prevScreenPosition.w)*0.5f+0.5f;
	velocity = (screenPosition - prevScreenPosition)*100.f+0.5f;

	vec3 diffuseLight = accumLight.xyz;
	vec3 specularLight = diffuseLight * accumLight.w / (lum(diffuseLight)+0.00001f);

	outColor = diffuseColor * (
		ambientLight + 
		diffuseLight * reflectionFactors.x +
		specularFactor * (
			specularLight * reflectionFactors.y +
			environmentColor * reflectionFactors.z
		)
	);
}