#version 430 core

#define %shadowing_technique% // either PCF or PCSS

noperspective in vec2 vf_ndcCoordinates;
flat in int vf_instanceID;

uniform sampler2D normalMap; //slot 0, these values are normalized from [-1,1] -> [0,1]
uniform sampler2D depthMap; //slot 1, stores the z-coordinate of the view position (is always negative)
uniform sampler2D shininessMap; // slot 2
uniform sampler2D perPixelNormalMap; // slot 3, these values are normalized from [-1,1] -> [0,1]
uniform sampler2DArrayShadow shadowMapArrayCompare; // slot 4
#ifdef PCSS
	uniform sampler2DArray shadowMapArray; // slot 5
	
	layout(std140, binding=2) uniform DepthConversionBlock {
		vec4 depthConversion[100];
	};

	flat in int vf_depthConversionID;
#endif

// stay constant over longer period of time
uniform vec2 projectionEntries;

struct SpotLightShadowProperties {
	mat4 viewToShadowMatrix;
	vec3 lightViewPosition;
	vec3 lightViewNegatedDirection;
	vec3 lightColor;
	vec4 lightSpotFactors_shadowMapIndex; // from d3d9 model: cos(phi/2), cos(theta/2)-cos(phi/2), falloff
	vec4 lightAttenuation_radius_c_biasfactor; // 1/(range^2), range, (2/width)*tan(phi/2), 0.5*light_p3-0.5
};

// different for each light
layout(std140, binding=1) uniform SpotLightShadowPropertiesBlock {
	SpotLightShadowProperties prop[100];
};


layout(location=0) out vec4 accumulatedLight;

#include <lum.glsl>
#ifdef PCSS
	#include <poisson_disk.glsl>
#endif

void main(void)
{
	ivec2 screenCoords = ivec2(gl_FragCoord.xy);

	// view position from depth and screen position
	vec4 surfaceViewPosition;
	surfaceViewPosition.zw = vec2(texelFetch(depthMap, screenCoords, 0).x, 1.f);
	surfaceViewPosition.xy = -vf_ndcCoordinates * surfaceViewPosition.z / projectionEntries.xy;

	vec3 l = prop[vf_instanceID].lightViewPosition - surfaceViewPosition.xyz; // from surface point to light source
	float dist = length(l);

	vec4 lightAttenuation_radius_c_biasfactor = prop[vf_instanceID].lightAttenuation_radius_c_biasfactor;

	if (dist < lightAttenuation_radius_c_biasfactor.y)
	{
		// get variables
		l = l / dist;
		vec3 n = 2.f * texelFetch(normalMap, screenCoords, 0).xyz - 1.f;
		vec3 ppn = 2.f * texelFetch(perPixelNormalMap, screenCoords, 0).xyz - 1.f;
		
		vec3 v = normalize(surfaceViewPosition.xyz); // from viewer to surface point (negated view vector)

		// spot
		float rho = dot(l, prop[vf_instanceID].lightViewNegatedDirection); // cos of angle between l and lightdirection
		vec4 lightSpotFactors_shadowMapIndex = prop[vf_instanceID].lightSpotFactors_shadowMapIndex;
		float spot = pow(
			clamp( (rho - lightSpotFactors_shadowMapIndex.x) / lightSpotFactors_shadowMapIndex.y, 0.0f, 1.0f ),
			lightSpotFactors_shadowMapIndex.z );

		// diffuse
		float cosTheta = max(0.f, dot(n,l));
		float ppCosTheta = dot(ppn, l);

		// specular
		float shininess = texelFetch(shininessMap, screenCoords, 0).x * 255.f;
		vec3 r = reflect(l,n); // negated reflection vector
		float specular = max(0.f, pow(dot(v,r), shininess)); // both v and r are negated -> dot product is correct
		float visible = 1.f;

		#if (defined PCF || defined PCSS)
			// shadow
			vec4 shadowCoordinates = prop[vf_instanceID].viewToShadowMatrix * surfaceViewPosition;
			// width of depth map texel projected to depth of this pixel
			float c = shadowCoordinates.w * lightAttenuation_radius_c_biasfactor.z;
			// first term = tan(acos(cosTheta), first two terms are maximum needed pushback, third term converts to homogeneous depth bias space
			float clampedCosTheta = clamp(ppCosTheta,0.001f,0.999f);
			float angularBias = min(2.f,(sqrt(1-clampedCosTheta*clampedCosTheta)/clampedCosTheta)) * 1.42f * c;
			//float angularBias = tan(acos(clampedCosTheta)) * 1.1f;

			#if (defined PCF)
				float bias = 0.0f;
				visible = 
					texture(shadowMapArrayCompare, vec4(
						shadowCoordinates.xy / shadowCoordinates.w,
						lightSpotFactors_shadowMapIndex.w,
						(shadowCoordinates.z+bias)/(shadowCoordinates.w-bias))
					);
			#elif (defined PCSS)
				float searchRadius = 12.f / 512.f;
				vec3 textureShadowCoordinates = shadowCoordinates.xyz / shadowCoordinates.w;
				float shadowMapIndex = lightSpotFactors_shadowMapIndex.w;
				float biasFactor = prop[vf_instanceID].lightAttenuation_radius_c_biasfactor.w;
				float numBlockers = 0;
				float avgBlockerDepth = 0.f;
				vec4 depthConv = depthConversion[vf_depthConversionID];
			
				for (int i=0; i<16; i++)
				{
					vec2 offset = poissonDisk1[i] * searchRadius;
					float bias = (angularBias) * length(offset) * 512.f * 1.5f * biasFactor - 0.002f;
					float compare = (shadowCoordinates.z+bias)/(shadowCoordinates.w-bias);
					float shadowMapDepth = texture(shadowMapArray, vec3(textureShadowCoordinates.xy + offset, shadowMapIndex));
					float linearDepth = (depthConv.x + depthConv.y * shadowMapDepth) / (depthConv.z + depthConv.w * shadowMapDepth);
					float blocking = step(shadowMapDepth, compare);

					avgBlockerDepth += blocking * linearDepth;
					numBlockers += blocking;
				}

				//visible = 0.5f+numBlockers;
				
				if (numBlockers > 0.f)
				{
					avgBlockerDepth /= numBlockers;
					float linearReceiverDepth = (depthConv.x + depthConv.y * textureShadowCoordinates.z) / (depthConv.z + depthConv.w * textureShadowCoordinates.z);
					float filterRadius = min(5.f/512.f, (0.2f+2.f*(linearReceiverDepth - avgBlockerDepth)) * searchRadius);
					float sum = 0.0f;
					for (int i=0; i<32; i++)
					{
						vec2 offset = poissonDisk2[i] * filterRadius;
						float bias = (angularBias) * length(offset) * 512.f * 1.5f * biasFactor - 0.002f;
						float compare = (shadowCoordinates.z+bias)/(shadowCoordinates.w-bias);
						float value = 
							texture(shadowMapArrayCompare, vec4(
								textureShadowCoordinates.xy+offset,
								shadowMapIndex,
								compare)
							);
						sum += value;
					}
					visible = sum / 32.f;
				}
				
			#endif
		#endif


		// attenuate and store
		float attenuation = 1.f - (dist*dist)*lightAttenuation_radius_c_biasfactor.x;

		float backfaceFactor = clamp(10.f*(ppCosTheta+0.1f), 0.f, 1.f);
		vec3 attenuatedColor = prop[vf_instanceID].lightColor * attenuation * attenuation * cosTheta * spot * visible * backfaceFactor;

		accumulatedLight.xyz = attenuatedColor;
		accumulatedLight.w = lum(attenuatedColor * specular);
	}
	else
	{
		accumulatedLight = vec4(0.f,0.f,0.f,0.f);
	}
}