#version 330 core

out vec4 outColor;

uniform sampler2D depthMap; // perspective projection depth
uniform sampler2D helperLookup; // F(u,v) = int_0^v (exp(-u*tan(x)) dx)
uniform vec4 scale_offset;

uniform float distanceScale;

uniform vec4 projectionEntries;
uniform vec2 beta_intensity;
uniform vec2 invImageSize;

uniform vec3 viewLightPosition;

in float lightViewDepth;

void main(void)
{
	// if light volume surface point is behind closest geometry surface point, project on geometry surface point
	float depth = texelFetch(depthMap, ivec2(gl_FragCoord.xy), 0).r;
	float projectedDepth = min(gl_FragCoord.z, depth);

	// unproject updated light volume surface point to camera view space
	float ndcDepth = projectedDepth*2.f-1.f;
	float viewDepth = -projectionEntries.w / (projectionEntries.z + ndcDepth);
	vec2 ndcXY = (gl_FragCoord.xy+0.5f)*invImageSize*2.f - 1.f;
	vec3 viewSurfacePoint = vec3( -ndcXY*viewDepth / projectionEntries.xy, viewDepth);

	vec3 vLPScale = viewLightPosition * distanceScale;
	vec3 vSPScale = viewSurfacePoint * distanceScale;

	// the following formulae are derived in "A Practical Analytic Single Scattering Model for Real Time Rendering" by Sun et al.
	// helperLookup contains the Lookup table F(u,v) = int_0^v (exp(-u*tan(x)) dx)
	float beta = beta_intensity.x;
	float I0 = beta_intensity.y;

	float D_sv = length(vLPScale);
	float T_sv = beta * D_sv;
	float D_vp = length(vSPScale);
	float T_vp = beta * D_vp;

	float cosGamma = dot(vLPScale, vSPScale) / (D_sv*D_vp);
	float gamma = acos(cosGamma);
	float sinGamma = sin(gamma);

	float T_svTimesCosGamma = T_sv * cosGamma;
	float T_svTimesSinGamma = T_sv * sinGamma;
	
	float A0 = (beta * beta * I0 * exp(-T_svTimesCosGamma)) / (2.0*3.1415927 * T_svTimesSinGamma);
	float A1 = T_sv * sinGamma;
	float v = 3.1415927 / 4.0 + 0.5*atan((T_vp - T_svTimesCosGamma) / T_svTimesSinGamma);

	float f1 = texture2D(helperLookup, vec2(A1, v)*scale_offset.xy+scale_offset.zw).r;
	float f2 = texture2D(helperLookup, vec2(A1, gamma/2)*scale_offset.xy+scale_offset.zw).r;
	float airlight = A0 * (f1-f2);

	float sgn = mix(1.f, -1.f, gl_FrontFacing);// add back faces, subtract front faces
	float strength = airlight * sgn;
	
	//strength = mix(sgn*5.f, strength, strength==strength);
	outColor = vec4(strength,strength,strength,1.f);
}